import abc
import os
from pathlib import Path
from shutil import copy2
from typing import Generator, List

import setuptools.command.sdist


class GitSdistABC(abc.ABC, setuptools.command.sdist.sdist):
    """
    This class defines a custom command to build source distribution. It covers the case
    where projects store the setup.cfg file in a subfolder and are git repositories.

    The problem in this setup is that only the subfolder by default is packaged
    in the sdist, resulting in a archive that only contains part of the whole repo.
    """

    def make_release_tree(self, base_dir, files) -> None:
        """
        This method is the responsible of building the list of files that are included
        in the source distribution.

        These files (that could be the entire git tree) are copied in the subfolder
        before creating the archive, resulting in a source distribution like if the
        setup.cfg would be part of the top-level folder.

        This method prepares the egg metadata for the source distribution and allows
        specifying a list of files that are copied in the location of the setup.cfg.
        """

        import setuptools_scm.integration

        # Build the setuptools_scm configuration, containing useful info for the sdist
        config: setuptools_scm.integration.Configuration = (
            setuptools_scm.integration.Configuration.from_file(
                dist_name=self.distribution.metadata.name
            )
        )

        # Get the root of the git repository
        repo_root = config.absolute_root

        if not Path(repo_root).exists() or not Path(repo_root).is_dir():
            raise RuntimeError(f"Failed to find a git repo in {repo_root}")

        # Prepare the release tree by calling the original method
        super(GitSdistABC, self).make_release_tree(base_dir=base_dir, files=files)

        # Collect all the files and copy them in the subfolder containing setup.cfg
        for file in self.get_sdist_files(repo_root=repo_root):

            src = Path(file)
            dst = Path(base_dir) / Path(file).relative_to(repo_root)

            # Make sure that the parent directory of the destination exists
            dst.absolute().parent.mkdir(parents=True, exist_ok=True)

            print(f"{Path(file).relative_to(repo_root)} -> {dst}")
            copy2(src=src, dst=dst)

        # Create the updated list of files included in the sdist
        all_files_gen = Path(base_dir).glob(pattern="**/*")
        all_files = [str(f.relative_to(base_dir)) for f in all_files_gen]

        # Find the SOURCES.txt file
        sources_txt_list = list(Path(base_dir).glob(pattern=f"*.egg-info/SOURCES.txt"))
        assert len(sources_txt_list) == 1

        # Update the SOURCES.txt files with the real content of the sdist
        os.unlink(sources_txt_list[0])
        with open(file=sources_txt_list[0], mode="w") as f:
            f.write("\n".join([str(f) for f in all_files]))

    @staticmethod
    @abc.abstractmethod
    def get_sdist_files(repo_root: str) -> List[Path]:
        """
        Return all files that will be included in the source distribution.

        Note: the egg metadata are included by default, there's not need to specify them.

        Args:
            repo_root: The path to the root of the git repository.

        Returns:
            The list of files to include in the source distribution.
        """


class GitSdistTree(GitSdistABC):
    """
    This class defines a custom command to build source distribution. It covers the case
    where projects store the setup.cfg file in a subfolder and are git repositories.

    The problem in this setup is that only the subfolder by default is packaged
    in the sdist, resulting in a archive that only contains part of the whole repo.

    In particular, this class copies all the files that are part of the last git commit.
    Any uncommitted or staged files are ignored and are not part of the sdist.
    """

    @staticmethod
    def get_sdist_files(repo_root: str) -> List[Path]:

        import git

        # Create the git Repo object
        git_repo = git.Repo(path=repo_root)

        # Get all the files of the git repo recursively
        def list_paths(
            tree: git.Tree, path: Path = Path(".")
        ) -> Generator[Path, None, None]:

            for blob in tree.blobs:
                yield path / blob.name

            for tree in tree.trees:
                yield from list_paths(tree=tree, path=path / tree.name)

        # Return the list of absolute paths to all the git repo files
        return list(list_paths(tree=git_repo.commit().tree, path=Path(repo_root)))


class GitSdistFolder(GitSdistABC):
    """
    This class defines a custom command to build source distribution. It covers the case
    where projects store the setup.cfg file in a subfolder and are git repositories.

    The problem in this setup is that only the subfolder by default is packaged
    in the sdist, resulting in a archive that only contains part of the whole repo.

    In particular, this class copies all the files that are part of the git folder.
    It includes also all uncommitted and staged files.
    """

    @staticmethod
    def get_sdist_files(repo_root: str) -> List[Path]:

        # Create the list of files of the git folder
        all_files_gen = Path(repo_root).glob(pattern="**/*")

        # Return the list of absolute paths to all the git folder files (also uncommited)
        return [f for f in all_files_gen if not f.is_dir() and ".git" not in f.parts]



import inspect
import os
import sys
from pathlib import Path

import cmake_build_extension
import setuptools

# Importing the bindings inside the build_extension_env context manager is necessary only
# in Windows with Python>=3.8.
# See https://github.com/diegoferigo/cmake-build-extension/issues/8.
# Note that if this manager is used in the init file, cmake-build-extension becomes an
# install_requires that must be added to the setup.cfg. Otherwise, cmake-build-extension
# could only be listed as build-system requires in pyproject.toml since it would only
# be necessary for packaging and not during runtime.
init_py = inspect.cleandoc(
    """
    import cmake_build_extension

    with cmake_build_extension.build_extension_env():
        from . import bindings
    """
)

# Extra options passed to the CI/CD pipeline that uses cibuildwheel
CIBW_CMAKE_OPTIONS = []
if "CIBUILDWHEEL" in os.environ and os.environ["CIBUILDWHEEL"] == "1":

    # The manylinux variant runs in Debian Stretch and it uses lib64 folder
    if sys.platform == "linux":
        CIBW_CMAKE_OPTIONS += ["-DCMAKE_INSTALL_LIBDIR=lib"]

    # Eigen is not found when installed with vcpkg because we don't pass the toolchain.
    # Passing directly the right folder as workaround. This is meant to work just in CI.
    if os.name == "nt":
        CIBW_CMAKE_OPTIONS += [
            "-DEigen3_DIR:PATH=C:/vcpkg/packages/eigen3_x64-windows/share/eigen3",
        ]

# This example is compliant with PEP517 and PEP518. It uses the setup.cfg file to store
# most of the package metadata. However, build extensions are not supported and must be
# configured in the setup.py.
setuptools.setup(
    # The resulting "mymath" archive contains two packages: mymath_swig and mymath_pybind.
    # This approach separates the two bindings types, typically just one of them is used.
    ext_modules=[
        cmake_build_extension.CMakeExtension(
            # This could be anything you like, it is used to create build folders
            name="SwigBindings",
            # Name of the resulting package name (import mymath_swig)
            install_prefix="mymath_swig",
            # Exposes the binary print_answer to the environment.
            # It requires also adding a new entry point in setup.cfg.
            expose_binaries=["bin/print_answer"],
            # Writes the content to the top-level __init__.py
            write_top_level_init=init_py,
            # Selects the folder where the main CMakeLists.txt is stored
            # (it could be a subfolder)
            source_dir=str(Path(__file__).parent.absolute()),
            cmake_configure_options=[
                # This option points CMake to the right Python interpreter, and helps
                # the logic of FindPython3.cmake to find the active version
                f"-DPython3_ROOT_DIR={Path(sys.prefix)}",
                "-DCALL_FROM_SETUP_PY:BOOL=ON",
                "-DBUILD_SHARED_LIBS:BOOL=OFF",
                # Select the bindings implementation
                "-DEXAMPLE_WITH_SWIG:BOOL=ON",
                "-DEXAMPLE_WITH_PYBIND11:BOOL=OFF",
            ]
            + CIBW_CMAKE_OPTIONS,
        ),
        cmake_build_extension.CMakeExtension(
            name="Pybind11Bindings",
            # Name of the resulting package name (import mymath_pybind11)
            install_prefix="mymath_pybind11",
            # Note: pybind11 is a build-system requirement specified in pyproject.toml,
            #       therefore pypa/pip or pypa/build will install it in the virtual
            #       environment created in /tmp during packaging.
            #       This cmake_depends_on option adds the pybind11 installation path
            #       to CMAKE_PREFIX_PATH so that the example finds the pybind11 targets
            #       even if it is not installed in the system.
            cmake_depends_on=["pybind11"],
            # Exposes the binary print_answer to the environment.
            # It requires also adding a new entry point in setup.cfg.
            expose_binaries=["bin/print_answer"],
            # Writes the content to the top-level __init__.py
            write_top_level_init=init_py,
            # Selects the folder where the main CMakeLists.txt is stored
            # (it could be a subfolder)
            source_dir=str(Path(__file__).parent.absolute()),
            cmake_configure_options=[
                # This option points CMake to the right Python interpreter, and helps
                # the logic of FindPython3.cmake to find the active version
                f"-DPython3_ROOT_DIR={Path(sys.prefix)}",
                "-DCALL_FROM_SETUP_PY:BOOL=ON",
                "-DBUILD_SHARED_LIBS:BOOL=OFF",
                # Select the bindings implementation
                "-DEXAMPLE_WITH_SWIG:BOOL=OFF",
                "-DEXAMPLE_WITH_PYBIND11:BOOL=ON",
            ]
            + CIBW_CMAKE_OPTIONS,
        ),
    ],
    cmdclass=dict(
        # Enable the CMakeExtension entries defined above
        build_ext=cmake_build_extension.BuildExtension,
        # If the setup.py or setup.cfg are in a subfolder wrt the main CMakeLists.txt,
        # you can use the following custom command to create the source distribution.
        # sdist=cmake_build_extension.GitSdistFolder
    ),
)



import sys

import mymath_pybind11.bindings as mymath
import numpy as np
import pytest


def test_dot():

    v1 = [1.0, 2, 3, -5.5, 42]
    v2 = [-3.2, 0, 13, 6, -3.14]

    result = mymath.dot(vector1=v1, vector2=v2)
    assert pytest.approx(result) == np.dot(v1, v2)


def test_normalize():

    v = [1.0, 2, 3, -5.5, 42]

    result = mymath.normalize(data=v)
    assert pytest.approx(result) == np.array(v) / np.linalg.norm(v)


def test_dot_numpy():

    v1 = np.array([1.0, 2, 3, -5.5, 42])
    v2 = np.array([-3.2, 0, 13, 6, -3.14])

    result = mymath.dot(vector1=v1, vector2=v2)
    assert pytest.approx(result) == np.dot(v1, v2)


def test_normalize_numpy():

    v = np.array([1.0, 2, 3, -5.5, 42])

    result = mymath.normalize_numpy(data=v)

    assert isinstance(result, np.ndarray)
    assert pytest.approx(result) == np.array(v) / np.linalg.norm(v)


@pytest.mark.skipif(
    sys.platform == "win32", reason="fails with the error: access violation"
)
def test_assertion():

    v1 = np.array([1.0, 2, 3, -5.5])
    v2 = np.array([42.0])

    with pytest.raises(RuntimeError):
        _ = mymath.dot(vector1=v1, vector2=v2)


@pytest.mark.skipif(
    sys.version_info < (3, 7), reason="capture_output and text require Python 3.7"
)
def test_executable():

    import subprocess

    result = subprocess.run("print_answer_pybind11", capture_output=True, text=True)
    assert result.stdout.strip() == "42"

    result = subprocess.run(
        [sys.executable, "-m", "mymath_pybind11.bin", "print_answer_pybind11"],
        capture_output=True,
        text=True,
    )
    assert result.stdout.strip() == "42"



from pathlib import Path
from typing import List

from setuptools import Extension


class CMakeExtension(Extension):
    """
    Custom setuptools extension that configures a CMake project.

    Args:
        name: The name of the extension.
        install_prefix: The path relative to the site-package directory where the CMake
            project is installed (typically the name of the Python package).
        disable_editable: Skip this extension in editable mode.
        write_top_level_init: Create a new top-level ``__init__.py`` file in the install
            prefix and write content.
        cmake_configure_options: List of additional CMake configure options (-DBAR=FOO).
        cmake_build_options: List of additional build-time arguments.
        source_dir: The location of the main CMakeLists.txt.
        cmake_build_type: The default build type of the CMake project.
        cmake_component: The name of component to install. Defaults to all components.
        cmake_depends_on: List of dependency packages containing required CMake projects.
        expose_binaries: List of binary paths to expose, relative to top-level directory.
    """

    def __init__(
        self,
        name: str,
        install_prefix: str = "",
        disable_editable: bool = False,
        write_top_level_init: str = None,
        cmake_configure_options: List[str] = (),
        cmake_install_options: List[str] = [],
        cmake_build_options: List[str] = [],
        source_dir: str = str(Path(".").absolute()),
        cmake_build_type: str = "Release",
        cmake_component: str = None,
        cmake_depends_on: List[str] = (),
        expose_binaries: List[str] = (),
    ):

        super().__init__(name=name, sources=[])

        if not Path(source_dir).is_absolute():
            source_dir = str(Path(".").absolute() / source_dir)

        if not Path(source_dir).absolute().is_dir():
            raise ValueError(f"Directory '{source_dir}' does not exist")

        self.install_prefix = install_prefix
        self.cmake_build_type = cmake_build_type
        self.disable_editable = disable_editable
        self.write_top_level_init = write_top_level_init
        self.cmake_depends_on = cmake_depends_on
        self.source_dir = str(Path(source_dir).absolute())
        self.cmake_configure_options = cmake_configure_options
        self.cmake_build_options = cmake_build_options
        self.cmake_install_options = cmake_install_options
        self.cmake_component = cmake_component
        self.expose_binaries = expose_binaries



import os
from contextlib import contextmanager
from pathlib import Path

from . import build_ext_option, sdist_command
from .build_extension import BuildExtension
from .cmake_extension import CMakeExtension
from .sdist_command import GitSdistFolder, GitSdistTree


@contextmanager
def build_extension_env():
    """
    Creates a context in which build extensions can be imported.

    It fixes a change of behaviour of Python >= 3.8 in Windows:
    https://docs.python.org/3/whatsnew/3.8.html#bpo-36085-whatsnew

    Other related resources:

    - https://stackoverflow.com/a/23805306
    - https://www.mail-archive.com/dev@subversion.apache.org/msg40414.html

    Example:

    .. code-block:: python

        from cmake_build_extension import build_extension_env

        with build_extension_env():
            from . import bindings
    """

    cookies = []

    # Windows and Python >= 3.8
    if hasattr(os, "add_dll_directory"):

        for path in os.environ.get("PATH", "").split(os.pathsep):

            if path and Path(path).is_absolute() and Path(path).is_dir():
                cookies.append(os.add_dll_directory(path))

    try:
        yield

    finally:
        for cookie in cookies:
            cookie.close()
